#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from enum import Enum
from django.conf import settings

logger = logging.getLogger("feishu")

# 环境变量名
DEFAULTS = {
    'FEISHU_TOKEN_EXPIRE_TIME': 7200,  # token时效, https://open.feishu.cn/document/ukTMukTMukTM/uIjNz4iM2MjLyYzM
    'FEISHU_TOKEN_UPDATE_TIME': 600,  # token提前更新的时间
    'FEISHU_BATCH_SEND_SIZE': 200,  # 批量发送消息列表的大小限制
    'FEISHU_APP_ID': None,
    'FEISHU_APP_SECRET': None,
    'FEISHU_VERIFY_TOKEN': None,
    'FEISHU_ENCRYPT_KEY': None
}


class APISettings:
    def __init__(self, defaults=None):
        self.defaults = defaults or DEFAULTS
        self.user_settings = getattr(settings, 'FEISHU_V2_SDK_CONFIGS', None)  # CKEditor 5 configuration in settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]
        # Cache the result
        setattr(self, attr, val)
        return val


api_settings = APISettings(DEFAULTS)


# try:
#     if settings.FEISHU_APP_ID:
#         FEISHU_APP_ID = settings.FEISHU_APP_ID
#     if settings.FEISHU_APP_SECRET:
#         FEISHU_APP_SECRET = settings.FEISHU_APP_SECRET
#     if settings.FEISHU_ENCRYPT_KEY:
#         FEISHU_ENCRYPT_KEY = settings.FEISHU_ENCRYPT_KEY
#     if settings.FEISHU_VERIFY_TOKEN:
#         FEISHU_VERIFY_TOKEN = settings.FEISHU_VERIFY_TOKEN
#     if settings.FEISHU_TOKEN_EXPIRE_TIME:
#         FEISHU_TOKEN_EXPIRE_TIME = settings.FEISHU_TOKEN_EXPIRE_TIME
#     if settings.FEISHU_TOKEN_UPDATE_TIME:
#         FEISHU_TOKEN_UPDATE_TIME = settings.FEISHU_TOKEN_UPDATE_TIME
#     if settings.FEISHU_BATCH_SEND_SIZE:
#         FEISHU_BATCH_SEND_SIZE = settings.FEISHU_BATCH_SEND_SIZE
# except Exception as e:
#     logger.debug(f"导入飞书应用和事件的配置出现报错(%s)，请查看settings配置" % e)


class AppType(str, Enum):
    TENANT = "tenant"  # 企业自建应用
    USER = "user"  # 第三方应用



#!/usr/bin/env python
# coding: utf-8 -*-
from enum import Enum


class FeishuError(Exception):
    def __init__(self, code: int, msg: str):
        self.code = code
        self.msg = msg

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.code},{self.msg}>"

    __unicode__ = __repr__


class ERRORS(int, Enum):
    FAILED_TO_ESTABLISH_CONNECTION = -1
    UNABLE_TO_PARSE_SERVER_RESPONSE = -2
    UNKNOWN_SERVER_ERROR = -3
    UNSUPPORTED_METHOD = -4
    INVALID_IMAGE_FILE_OR_CONTENT = -5
    VALIDATION_ERROR = -6
    MISSING_ENCRYPT_KEY = -7



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from enum import Enum
from typing import List, Union, Optional

from pydantic import BaseModel


class SendMsgType(str, Enum):
    TEXT = "text"
    IMAGE = "image"
    POST = "post"
    SHARE_CHAT = "share_chat"
    INTERACTIVE = "interactive"


class PostTag(str, Enum):
    TEXT = "text"
    A = "a"
    AT = "at"
    IMG = "img"


class Content(BaseModel):
    pass


class TextContent(Content):
    text: str


class ImageContent(Content):
    image_key: str


class ShareChatContent(Content):
    share_chat_id: str


class PostElement(BaseModel):
    tag: PostTag


class PostTextElement(PostElement):
    """文本元素
    {
        "tag": "text",
        "un_escape": true,
        "text": "第一行&nbsp;:"
    }
    """
    tag: PostTag = PostTag.TEXT
    text: str
    un_escape: Optional[bool] = None


class PostAElement(PostElement):
    """超链接元素
    {
        "tag": "a",
        "text": "超链接",
        "href": "http://www.feishu.cn"
    }
    """
    tag: PostTag = PostTag.A
    text: str
    href: str
    un_escape: Optional[bool] = None


class PostAtElement(PostElement):
    """AT元素

    user_id可以是open_id/user_id
    {
        "tag": "at",
        "user_id": "ou_18eac85d35a26f989317ad4f02e8bbbb"
    }
    """
    tag: PostTag = PostTag.AT
    user_id: str


class PostImgElement(PostElement):
    """图片元素

    {
        "tag": "img",
        "image_key": "d640eeea-4d2f-4cb3-88d8-c964fab53987",
        "width": 300,
        "height": 300
    }
    """
    tag: PostTag = PostTag.IMG
    image_key: str
    height: int
    width: int


class I18nPost(BaseModel):
    """Post内容

    https://open.feishu.cn/document/ukTMukTMukTM/uMDMxEjLzATMx4yMwETM

    效果:
    https://sf1-ttcdn-tos.pstatp.com/obj/website-img/a3074630f25fb0e7d3df394c1348ba41_cn.png
    {
        "title": "我是一个标题",
        "content": [
            [
                {
                    "tag": "text",
                    "un_escape": true,
                    "text": "第一行&nbsp;:"
                },
                {
                    "tag": "a",
                    "text": "超链接",
                    "href": "http://www.feishu.cn"
                },
                {
                    "tag": "at",
                    "user_id": "ou_18eac85d35a26f989317ad4f02e8bbbb"
                }
            ],
            [
                {
                    "tag": "text",
                    "text": "第二行 :"
                },
                {
                    "tag": "text",
                    "text": "文本测试"
                }
            ],
            [
                {
                    "tag": "img",
                    "image_key": "d640eeea-4d2f-4cb3-88d8-c964fab53987",
                    "width": 300,
                    "height": 300
                }
            ]
        ]
    }
    """
    title: Optional[str] = None
    content: List[List[PostElement]]


class Post(BaseModel):
    """ 这里暂未处理i18n相关, 其实可以都列一下 """
    zh_cn: Optional[Union[I18nPost, dict]] = None
    en_us: Optional[Union[I18nPost, dict]] = None


class PostContent(Content):
    post: Post


class Message(BaseModel):
    """消息基类

    Args:
        chat_id 同下
        open_id 同下
        user_id 同下
        email 如下
            - 私聊的话open_id/user_id/email三个里面要填一个
            - 群聊的话必须填chat_id
            - 如果都提供的话，优先级为 chat_id > open_id > user_id > email
        root_id 回复消息时，对应消息的消息ID，选填
        msg_type 消息类型, 见MsgType
        content 消息内容, 见Content
    """
    chat_id: Optional[str] = None
    open_id: Optional[str] = None
    user_id: Optional[str] = None
    email: Optional[str] = None
    root_id: Optional[str] = None
    msg_type: SendMsgType
    content: Content


class TextMessage(Message):
    """文本消息

    {
       "open_id":"ou_5ad573a6411d72b8305fda3a9c15c70e",
       "root_id":"om_40eb06e7b84dc71c03e009ad3c754195",
       "chat_id":"oc_5ad11d72b830411d72b836c20",
       "user_id": "92e39a99",
       "email":"fanlv@gmail.com",
       "msg_type":"text",
       "content":{
            "text":"text content<at user_id=\"ou_88a56e7e8e9f680b682f6905cc09098e\">test</at>"
        }
    }
    """
    msg_type: SendMsgType = SendMsgType.TEXT
    content: TextContent


class ImageMessage(Message):
    """图片消息

    {
       "open_id":"ou_5ad573a6411d72b8305fda3a9c15c70e",
       "chat_id":"oc_5ad11d72b830411d72b836c20",
       "root_id":"om_40eb06e7b84dc71c03e009ad3c754195",
        "user_id": "92e39a99",
        "email":"fanlv@gmail.com",
        "msg_type":"image",
        "content":{
            "image_key": "1a0c4cb9-c680-4371-924c-ddb5f2750c3d"
        }
    }
    """
    msg_type: SendMsgType = SendMsgType.IMAGE
    content: ImageContent


class PostMessage(Message):
    """富文本消息

    {
       "open_id":"ou_5ad573a6411d72b8305fda3a9c15c70e",
       "root_id":"om_40eb06e7b84dc71c03e009ad3c754195",
       "chat_id":"oc_5ad11d72b830411d72b836c20",
       "user_id": "92e39a99",
       "email":"fanlv@gmail.com",
       "msg_type":"post",
       "content":{
            "post":{
                "zh_cn":{}, // option
                "ja_jp":{}, // option
                "en_us":{} // option
            }
       }
    }
    """
    msg_type: SendMsgType = SendMsgType.POST
    content: PostContent


class ShareChatMessage(Message):
    """群名片

    {
       "open_id":"ou_5ad573a6411d72b8305fda3a9c15c70e",
       "root_id":"om_40eb06e7b84dc71c03e009ad3c754195",
       "chat_id":"oc_5ad11d72b830411d72b836c20",
       "user_id": "92e39a99",
        "msg_type": "share_chat",
        "content":{
            "share_chat_id": "oc_f5b1a7eb27ae2c7b6adc2a74faf339ff"
        }
    }
    """
    msg_type: SendMsgType = SendMsgType.SHARE_CHAT
    content: ShareChatContent


class ReadUser(BaseModel):
    """已读用户

    {
        "open_id": "ou_18eac85d35a26f989317ad4f02e8bbbb",
        "timestamp": "1570697776",
        "user_id": "ca51d83b"
    }
    """
    open_id: str
    timestamp: str
    user_id: Optional[str] = None


class BatchMessage(BaseModel):
    """批量消息

    {
        "department_ids": [
            "od-9d1912863dceba33a33226b52b32a776",
            "od-81d13d502aaa9514059ea570c37d6438"
        ],
        "open_ids": [
            "ou_18eac85d35a26f989317ad4f02e8bbbb",
            "ou_461cf042d9eedaa60d445f26dc747d5e"
        ],
        "user_ids": [
            "7cdcc7c2",
            "ca51d83b"
        ],
        "msg_type": "text",
        "content": {
            "text": "test content"
        }
    }
    """
    department_ids: List[str] = []
    open_ids: List[str] = []
    user_ids: List[str] = []
    msg_type: SendMsgType
    content: Content


class BatchSendResponse(BaseModel):
    """批量发送消息返回

    {
        "invalid_department_ids": [
            "od-d3079cfb6ed783b9bdc2e9eac04e85b41"
        ],
        "invalid_open_ids": [
            "ou_456e168d61cec276083b357f7bd3f1491",
            "ou_f8cbdb26fb2e4eda075e003381a102a41"
        ],
        "invalid_user_ids": ["7cdcc7c22"],
        "message_id": "bm-d4be107c616aed9c1da8ed8068570a9f"
    }
    """
    message_id: str
    message_ids: List[str] = []  # 拆分成多个请求的话会有多个message_id
    invalid_department_ids: List[str] = []
    invalid_open_ids: List[str] = []
    invalid_user_ids: List[str] = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""飞书订阅事件

参考订阅事件文档 https://open.feishu.cn/document/ukTMukTMukTM/uUTNz4SN1MjL1UzM
"""

from .base import EventContent
from .bot import *
from .contact import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import asyncio
import base64
import hashlib
import inspect
import json
import linecache
import logging
import os
import re
import secrets
import textwrap
from asyncio import AbstractEventLoop
from functools import wraps

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from ..baseclient import FeishuBaseClient
from ..settings import *
from ..errors import FeishuError, ERRORS
from ..models import *

# 需要models/typing/consts的import存在
__needs__ = [Union, EventContent, api_settings.FEISHU_APP_ID]


class BaseAPI:
    logger = logging.getLogger("feishu")

    def __init__(self):
        self.client: Optional[FeishuBaseClient] = None

    def adapt_sync_and_async(self, method, *args, **kwargs):
        pass


def verify_signature(verify_token: str, headers: dict, body: bytes) -> bool:
    """校验消息卡片签名

    https://open.feishu.cn/document/ukTMukTMukTM/uYzMxEjL2MTMx4iNzETM
    Args:
        verify_token: 飞书后台事件订阅中的Verification Token
        headers: 返回的HTTP Headers
        body: 返回的HTTP Body
    """
    timestamp = headers.get('X-Lark-Request-Timestamp')
    nonce = headers.get('X-Lark-Request-Nonce')
    encoded = (timestamp + nonce + verify_token).encode('utf-8')
    signature = hashlib.sha1(encoded + body).hexdigest()
    return headers['X-Lark-Signature'] == signature


def decrypt_aes(encrypt_key: str, encrypted: str) -> dict:
    """解密飞书事件回调加密部分

    https://open.feishu.cn/document/ukTMukTMukTM/uUTNz4SN1MjL1UzM#%E9%80%9A%E8%BF%87Encrypt%20Key%E5%8A%A0%E5%AF%86%E6%95%B0%E6%8D%AE
    Args:
        encrypt_key: 飞书后台加密用的encrypt_key
        encrypted: 加密的密文
    """
    if not encrypt_key:
        raise FeishuError(ERRORS.MISSING_ENCRYPT_KEY, "飞书推送了需要解密的消息, 但是配置的encrypt_key为空")
    block_size = 16
    decoded = base64.b64decode(encrypted)
    iv = decoded[:block_size]
    key = hashlib.sha256(encrypt_key.encode()).digest()
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
    raw = cipher.decryptor().update(decoded[block_size:])
    if raw[-1] <= block_size:
        # unpad
        raw = raw[:-raw[-1]]
    return json.loads(raw)


def _get_or_create_event_loop() -> AbstractEventLoop:
    try:
        loop = asyncio.get_event_loop()
        return loop
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop


to_be_created = {}
context = {}
# name -> newname
# "= name(...)" -> "= await newname(...)"
async_method_mapping = {
    "self.client.request": "self.client.request",
    "self.client.fetch": "self.client.fetch"}


def allow_async_call(func):
    """给同步方法加上被异步调用的能力

    为了让异步调用中不会被同步方法卡住event_loop
    加allow_async_call修饰的方法必须做到以下几点:

    - 方法中没有同步IO事件, 读写文件都最好不要有(本地磁盘且小文件问题不大)
    - API请求用self.client.request, 且一定要写成xxxx = self.client.request这样的形式
    - 通用HTTP请求用self.client.fetch, 且一定要携程yyyy = self.client.fetch这样的形式
    """
    global context, async_method_mapping, to_be_created
    name = func.__name__
    newname = name + "_async"
    async_method_mapping["self." + name] = "self." + newname

    def create_async_api(func=func, name=name):
        """生成async版本的API, 名字为原方法+'_async'

        Args:
            func: 同步版本的函数
            name: 同步版本的函数名
        """
        # 用正则生成一下async版本的代码
        # 注意，这个代码只在context中没有找到async代码时才会执行
        # 这里更好的方案是用ast，但是太麻烦了——正则大部分时候ok就行
        # def f(self, a):                             async f_async(self, a):
        #     ...                                         ...
        #     xxxx = self.client.request(...   =>         xxxx = await self.client.request(...
        #     yyyy = self.client.fetch(...     =>         yyyy = await self.client.fetch(...
        #     ...                                         ...
        source = inspect.getsource(func)
        source = textwrap.dedent(source)

        # 加上async def
        source2 = re.compile(r'def(\s*' + name + r')(\(.*?)->[^:]*:\s*\n    ', re.DOTALL).sub(
            r'async def\1_async\2-> "Future":\n    ', source)
        if source2 == source:
            source2 = re.compile(r'def(\s*' + name + r')(\(.*?)\):\s*\n    ', re.DOTALL).sub(
                r'async def\1_async\2) -> "Future":\n    ', source2)
        source = source2

        # 修改await, 目前就修改两种:
        # - xxx = self.method
        # - return self.method
        for method, newmethod in async_method_mapping.items():
            source = re.sub(r'    (\S+)\s*=\s*(' + re.escape(method) + r')(.*?\n(\n|    ))',
                            r'    \1 = await ' + newmethod + r'\3',
                            source)
            source = re.sub(r'    (\s*)return\s*(' + re.escape(method) + r')(.*?\n(\n|    ))',
                            r'    \1return await ' + newmethod + r'\3',
                            source)
        # 去掉decorator
        source = re.sub(r'\s*@allow_async_call\s*\n', '', source)

        # 试着从源文件import各种依赖，以免调用的时候出问题
        filename = func.__code__.co_filename[:-3].replace(os.path.sep, '.')
        paths = filename.split('.')
        for i in range(1, min(6, len(paths) + 1)):
            path = ".".join(paths[-i:])
            # if "." not in path:
            #     path = "." + path
            if "-" in path:
                break
            try:
                insert = f"from {path} import *\n"
                exec(insert)
            except:
                continue
            else:
                source = insert + source

                # 成功一次就应该够了
                break

        # 生成临时文件并预编译
        # 这个输出太多了暂时禁用
        # BaseAPI.logger.debug(f"自动生成的async版本API:{name} ---\n{source}")
        filename = "api_" + secrets.token_hex(4) + ".py"
        compiled = compile(source, filename, mode="exec")
        linecache.cache[filename] = (len(source), None, [line + '\n' for line in source.splitlines()], filename)
        exec(compiled, globals())

        context[newname] = globals()[newname]

    to_be_created[newname] = create_async_api

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        def create_async_apis():
            for newname2, create in to_be_created.items():
                create()
                setattr(self.__class__, newname2, context[newname2])

        if not hasattr(self.__class__, newname):
            create_async_apis()

        if not self.client.run_async or name == 'dummy':
            return func(self, *args, **kwargs)
        else:
            # print("calling", self, newname, "args", args, "kwargs", kwargs)
            if not self.client.event_loop or self.client.event_loop.is_closed():
                self.client.event_loop = _get_or_create_event_loop()

            if not hasattr(self, newname):
                create_async_apis()

            return asyncio.ensure_future(
                getattr(self, newname)(*args, **kwargs),
                loop=self.client.event_loop
            )

    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
from abc import ABC, abstractmethod
from typing import Optional

from .settings import api_settings


class TokenStore(ABC):
    @abstractmethod
    def get(self, key: str, value: str, expire: float = api_settings.FEISHU_TOKEN_EXPIRE_TIME):
        pass

    @abstractmethod
    def get(self, key: str):
        pass


class MemoryStore(TokenStore):
    """ 内存存储 """
    cache = {}
    timings = {}

    def set(self, key: str, value: str, expire: float = api_settings.FEISHU_TOKEN_EXPIRE_TIME):
        expire -= api_settings.FEISHU_TOKEN_UPDATE_TIME
        self.cache[key] = value
        self.timings[key] = time.time() + expire

    def get(self, key: str):
        expired_time = self.timings.get(key)
        if expired_time and expired_time < time.time():
            self.timings.pop(key, None)
            self.cache.pop(key, None)
        return self.cache.get(key)


class RedisStore(TokenStore):
    """ Redis存储 """

    def __init__(self, redis_url: Optional[str] = None):
        import redis
        if redis_url:
            self.client = redis.Redis.from_url(redis_url)
        else:
            self.client = redis.Redis()

    def set(self, key: str, value: str, expire: float = api_settings.FEISHU_TOKEN_EXPIRE_TIME):
        expire -= api_settings.FEISHU_TOKEN_UPDATE_TIME
        self.client.setex(key, value, expire)

    def get(self, key: str):
        self.client.get(key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .apis import *
from .client import FeishuClient
from .errors import FeishuError, ERRORS
from .models import *
from .stores import TokenStore, MemoryStore, RedisStore



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Bot类型"""
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel


class BotActivateStatus(int, Enum):
    """
     0: 初始化，租户待安装,
     1: 租户停用,
     2: 租户启用,
     3: 安装后待启用,
     4: 升级待启用,
     5: license过期停用,
     6: Lark套餐到期或降级停用,
    """
    INITIALIZED = 0
    DISABLED = 1
    ENABLED = 2
    INSTALLED = 3
    UPGRADED = 4
    LICENCE_EXPIRED = 5
    SET_EXPIRED = 6


class BotChatType(str, Enum):
    GROUP = "group"
    P2P = "p2p"


class BotInfo(BaseModel):
    """
    {
        "activate_status": 2,
        "app_name": "name",
        "avatar_url": "https://sf1-ttcdn-tos.pstatp.com/obj/lark.avatar/da5xxxx14b16113",
        "ip_white_list": [],
        "open_id": "ou_xxx"
    }
    """
    activate_status: BotActivateStatus
    app_name: str
    avatar_url: str
    ip_white_list: List[str]
    open_id: str


class I18Names(BaseModel):
    zh_cn: str
    en_us: str


class CreateChatRequest(BaseModel):
    """创建群请求
    {
        "name": "group name",
        "description": "group description",
        "user_ids": [
            "33417745",
            "cb93bdca"
        ],
        "open_ids": [
            "ou_4065981088f8ef67a504ba8bd6b24d85",
            "ou_111111111111111111111111111111111"
        ],
        "i18n_names": {
            "zh_cn": "zh_cn name",
            "en_us": "en_us name",
            "ja_jp": "ja_jp name"
        },
        "only_owner_add": false,
        "share_allowed": true,
        "only_owner_at_all": false,
        "only_owner_edit": false
    }
    """
    name: str = ''
    description: str = ''
    open_ids: List[str] = []
    user_ids: List[str] = []
    i18_names: I18Names
    only_owner_add: bool = False
    share_allowed: bool = True
    only_owner_at_all: bool = False
    only_owner_edit: bool = False


class CreateChatResponse(BaseModel):
    """创建群返回
    {
        "chat_id": "oc_4f65b883a624c59414157668c91637ab",
        "invalid_open_ids": [
            "ou_111111111111111111111111111111111"
        ],
        "invalid_user_ids": [
            "33417745"
        ]
    }
    """
    chat_id: str = ''
    invalid_open_ids: List[str] = []
    invalid_user_ids: List[str] = []


AddChatterResponse = CreateChatResponse
RemoveChatterResponse = CreateChatResponse


class ChatMember(BaseModel):
    open_id: str
    user_id: str


class ChatInfo(BaseModel):
    avatar: str = ''
    description: str = ''
    chat_id: str = ''
    name: str = ''
    owner_open_id: str = ''
    owner_user_id: str = ''

    # only available in get_chat_info API
    i18n_names: I18Names = {}
    members: List[ChatMember] = []
    type: BotChatType = "group"


class ChatPagination(BaseModel):
    """
    {
        "groups": [
            {
                "avatar": "http://p3.pstatp.com/origin/78c0000676df676a7f6e",
                "chat_id": "oc_9e9619b938c9571c1c3165681cdaead5",
                "description": "description1",
                "name": "test1",
                "owner_open_id": "ou_194911f90c43ec42d1ba0e93f22b8fb1",
                "owner_user_id": "ca51d83b"
            },
            {
                "avatar": "http://p4.pstatp.com/origin/dae10015cfb346541010",
                "chat_id": "oc_5ce6d572455d361153b7cb51da133945",
                "description": "description2",
                "name": "test2",
                "owner_open_id": "ou_194911f90c43ec42d1ba0e93f22b8fb1",
                "owner_user_id": "ca51d83b"
            }
        ],
        "has_more": false,
        "page_token": "0"
    }
    """
    has_more: bool = False
    page_token: str = '0'
    groups: List[ChatInfo] = []


class ChatUpdateRequest(BaseModel):
    """
    {
        "chat_id": "oc_020ff1d91a0295fc3961032768d41f39",
        "owner_user_id":"cb93bdca",
        "owner_open_id":"ou_4065981088f8ef67a504ba8bd6b24d85",
        "name":"group name",
        "i18n_names":{
            "zh_cn":"zh_cn name",
            "en_us":"en_us name",
            "ja_jp":"ja_jp name"
        },
        "only_owner_add": false,
        "share_allowed": true,
        "only_owner_at_all": false,
        "only_owner_edit": false
    }
    """
    chat_id: str
    owner_open_id: str = ''
    owner_user_id: str = ''
    name: str = ''
    i18n_names: I18Names = {}

    only_owner_add: Optional[bool] = None
    share_allowed: Optional[bool] = None
    only_owner_at_all: Optional[bool] = None
    only_owner_edit: Optional[bool] = None



#!/usr/bin/env python3.8
class Dict2Obj(dict):

    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        self[key] = value


class InvalidEventException(Exception):
    def __init__(self, error_info):
        self.error_info = error_info

    def __str__(self) -> str:
        return "Invalid event: {}".format(self.error_info)

    __repr__ = __str__


# class Obj(dict):
#     def __init__(self, d):
#         for a, b in d.items():
#             if isinstance(b, (list, tuple)):
#                 setattr(self, a, [Obj(x) if isinstance(x, dict) else x for x in b])
#             else:
#                 setattr(self, a, Obj(b) if isinstance(b, dict) else b)
#
#
# def dict_2_obj(d: dict):
#     return Obj(d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""机器人管理相关API

包括:
- 群管理
- 机器人管理
"""
from typing import List, Optional

from .base import BaseAPI, allow_async_call
from ..settings import api_settings
from ..models import BotInfo, CreateChatRequest, CreateChatResponse, ChatPagination, ChatInfo, ChatUpdateRequest, \
    AddChatterResponse, RemoveChatterResponse


class BotAPI(BaseAPI):
    @allow_async_call
    def get_bot_info(self) -> BotInfo:
        """获取自建应用的token

        https://open.feishu.cn/document/ukTMukTMukTM/uAjMxEjLwITMx4CMyETM
        返回示例:
            {
                "activate_status": 2,
                "app_name": "name",
                "avatar_url": "https://sf1-ttcdn-tos.pstatp.com/obj/lark.avatar/da5xxxx14b16113",
                "ip_white_list": [],
                "open_id": "ou_xxx"
            }
        """
        api = "/bot/v3/info/"
        result = self.client.request("GET", api=api)
        return BotInfo(**result.get("bot", {}))

    @allow_async_call
    def add_bot(self, chat_id: str):
        """拉机器人进群

        Args:
            chat_id: 群ID

        失败会raise FeishuError, (code和msg对应飞书平台的code/msg)
        """
        api = "/bot/v4/add"
        payload = {"chat_id": chat_id}
        self.client.request("POST", api=api, payload=payload)

    @allow_async_call
    def remove_bot(self, chat_id: str):
        """将机器人移出群

        Args:
            chat_id: 群ID

        失败会raise FeishuError, (code和msg对应飞书平台的code/msg)
        """
        api = "/bot/v4/remove"
        payload = {"chat_id": chat_id}
        self.client.request("POST", api=api, payload=payload)

    @allow_async_call
    def create_chat(self, name: str = '', description: str = '',
                    open_ids: List[str] = [], user_ids: List[str] = [], i18n_names: dict = {},
                    only_owner_add: bool = False, share_allowed: bool = True,
                    only_owner_at_all: bool = False, only_owner_edit: bool = False) -> CreateChatResponse:
        """创建群

        Args:
            name: 群名称, 可不填
            description: 群描述, 可不填
            open_ids: 最多200个, 自建应用使用
            user_ids: 最多200个, 第三方应用使用
            i18n_names: 举例 {”zh_cn": "中文", "en_us": "english"}
            only_owner_add: 是否仅群主可以添加人
            share_allowed: 是否允许分享群
            only_owner_at_all: 是否仅群主@all
            only_owner_edit: 是否仅群主可编辑群信息，群信息包括头像、名称、描述、公告

        Returns:
            CreateChatResponse:
                chat_id
                invalid_open_ids
                invalid_user_ids
        """
        api = "/chat/v4/create/"
        req = CreateChatRequest(name=name, description=description, open_ids=open_ids, user_ids=user_ids,
                                i18n_names=i18n_names, only_owner_add=only_owner_add, share_allowed=share_allowed,
                                only_owner_at_all=only_owner_at_all, only_owner_edit=only_owner_edit)
        payload = req.dict(exclude_defaults=True)
        result = self.client.request("POST", api=api, payload=payload)
        return CreateChatResponse(**result.get("data"))

    @allow_async_call
    def list_chat(self, page_size: int = 100, page_token: str = '') -> ChatPagination:
        """获取群列表

        Args:
            page_size: 最大为200
            page_token: 分页标记，第一次请求不填，表示从头开始遍历
                分页查询还有更多群时会同时返回新的 page_token, 下次遍历可采用该 page_token 获取更多群
        Returns:
            ChatPagination
                has_more: bool
                page_token: str
                groups: List[ChatInfo]
            示例
            {
                "groups": [
                    {
                        "avatar": "http://p3.pstatp.com/origin/78c0000676df676a7f6e",
                        "chat_id": "oc_9e9619b938c9571c1c3165681cdaead5",
                        "description": "description1",
                        "name": "test1",
                        "owner_open_id": "ou_194911f90c43ec42d1ba0e93f22b8fb1",
                        "owner_user_id": "ca51d83b"
                    },
                    {
                        "avatar": "http://p4.pstatp.com/origin/dae10015cfb346541010",
                        "chat_id": "oc_5ce6d572455d361153b7cb51da133945",
                        "description": "description2",
                        "name": "test2",
                        "owner_open_id": "ou_194911f90c43ec42d1ba0e93f22b8fb1",
                        "owner_user_id": "ca51d83b"
                    }
                ],
                "has_more": false,
                "page_token": "0"
            }
        """
        api = "/chat/v4/list"
        payload = {
            "page_size": str(page_size),
        }
        if page_token:
            payload["page_token"] = page_token
        result = self.client.request("POST", api=api, payload=payload)
        return ChatPagination(**result)

    @allow_async_call
    def list_chat_all(self) -> List[ChatInfo]:
        """获取所有群列表"""
        all_chats = []
        pag = self.list_chat(page_size=api_settings.FEISHU_BATCH_SEND_SIZE)
        all_chats.extend(pag.groups)
        while pag.has_more:
            pag = self.list_chat(page_size=api_settings.FEISHU_BATCH_SEND_SIZE, page_token=pag.page_token)
            all_chats.extend(pag.groups)
        return all_chats

    @allow_async_call
    def get_chat_info(self, chat_id: str) -> ChatInfo:
        """获取群信息"""
        api = "/chat/v4"
        params = {"chat_id": chat_id}
        result = self.client.request("GET", api=api, params=params)
        return ChatInfo(**result.get("data", {}))

    @allow_async_call
    def update_chat_info(self, chat_id: str,
                         owner_user_id: Optional[str] = None,
                         owner_open_id: Optional[str] = None,
                         name: Optional[str] = None,
                         i18n_names: Optional[dict] = None,
                         only_owner_add: Optional[bool] = None,
                         share_allowed: Optional[bool] = None,
                         only_owner_at_all: Optional[bool] = None,
                         only_owner_edit: Optional[bool] = None
                         ) -> str:
        """更新群信息"""
        api = "/chat/v4/update/"
        req = ChatUpdateRequest(chat_id=chat_id)
        for key in ["owner_user_id", "owner_open_id", "name", "i18n_names",
                    "only_owner_add", "share_allowed", "only_owner_at_all", "only_owner_edit"]:
            value = locals()[key]
            if value is not None:
                setattr(req, key, value)
        payload = req.dict(exclude_unset=True)
        result = self.client.request("POST", api=api, payload=payload)
        return result.get("data", {}).get("chat_id", "")

    @allow_async_call
    def add_chatter(self, chat_id: str, user_ids: List[str] = [], open_ids: List[str] = []) -> AddChatterResponse:
        """拉用户进群

        Args:
            chat_id: 群ID
            open_ids: 自建应用提供用户ID, 数量限制200
            user_ids: 第三方应用提供用户ID, 数量限制200

        Returns:
            AddChatterResponse:
                invalid_user_ids
                invalid_open_ids
        """
        api = "/chat/v4/chatter/add/"
        payload = {
            "chat_id": chat_id
        }
        if user_ids:
            payload["user_ids"] = user_ids
        if open_ids:
            payload["open_ids"] = open_ids
        result = self.client.request("POST", api=api, payload=payload)
        response = AddChatterResponse(result.get("data", {}))
        response.chat_id = chat_id
        return response

    @allow_async_call
    def add_chatter_all(self, chat_id: str, user_ids: List[str] = [], open_ids: List[str] = [],
                        slice_size: int = api_settings.FEISHU_BATCH_SEND_SIZE) -> AddChatterResponse:
        """拉全部用户进群, 不考虑200限制"""
        response = AddChatterResponse()
        while user_ids or open_ids:
            resp = self.add_chatter(chat_id, user_ids[:slice_size], open_ids[:slice_size])
            response.invalid_user_ids.extend(resp.invalid_user_ids)
            response.invalid_open_ids.extend(resp.invalid_open_ids)
            user_ids, open_ids = user_ids[slice_size:], open_ids[slice_size:]
        response.chat_id = chat_id
        return response

    @allow_async_call
    def remove_chatter(self, chat_id: str, user_ids: List[str] = [], open_ids: List[str] = []) -> RemoveChatterResponse:
        """移除用户出群

        Args:
            chat_id: 群ID
            open_ids: 自建应用提供用户ID, 数量限制200
            user_ids: 第三方应用提供用户ID, 数量限制200

        Returns:
            RemoveChatterResponse:
                invalid_user_ids
                invalid_open_ids
        """
        api = "/chat/v4/chatter/delete/"
        payload = {
            "chat_id": chat_id
        }
        if user_ids:
            payload["user_ids"] = user_ids
        if open_ids:
            payload["open_ids"] = open_ids
        result = self.client.request("POST", api=api, payload=payload)
        response = RemoveChatterResponse(result.get("data", {}))
        response.chat_id = chat_id
        return response

    @allow_async_call
    def remove_chatter_all(self, chat_id: str, user_ids: List[str] = [], open_ids: List[str] = [],
                           slice_size: int = api_settings.FEISHU_BATCH_SEND_SIZE) -> RemoveChatterResponse:
        """移除全部用户出群, 不考虑200限制"""
        response = RemoveChatterResponse()
        while user_ids or open_ids:
            resp = self.remove_chatter(chat_id, user_ids[:slice_size], open_ids[:slice_size])
            response.invalid_user_ids.extend(resp.invalid_user_ids)
            response.invalid_open_ids.extend(resp.invalid_open_ids)
            user_ids, open_ids = user_ids[slice_size:], open_ids[slice_size:]
        response.chat_id = chat_id
        return response

    @allow_async_call
    def disband_chat(self, chat_id: str):
        """解散群

        如果解散失败会raise FeishuError, code/msg为飞书返回的失败原因
        """
        api = "/chat/v4/disband"
        payload = {"chat_id": chat_id}
        self.client.request("POST", api=api, payload=payload)

    @allow_async_call
    def create_chat_all(self, name: str = '', description: str = '',
                        open_ids: List[str] = [], user_ids: List[str] = [], i18n_names: dict = {},
                        only_owner_add: bool = False, share_allowed: bool = True,
                        only_owner_at_all: bool = False, only_owner_edit: bool = False,
                        slice_size: int = api_settings.FEISHU_BATCH_SEND_SIZE) -> CreateChatResponse:
        """创建群

        和create_chat的区别是, create_chat的列表有200个的限制，这个没有
        先create_chat, 如果还没搞定就add_chatter_all
        """
        first_batch_open_ids, left_open_ids = open_ids[:slice_size], open_ids[slice_size:]
        first_batch_user_ids, left_user_ids = user_ids[:slice_size], user_ids[slice_size:]
        response = self.create_chat(name=name, description=description, i18n_names=i18n_names,
                                    open_ids=first_batch_open_ids, user_ids=first_batch_user_ids,
                                    only_owner_add=only_owner_add, share_allowed=share_allowed,
                                    only_owner_at_all=only_owner_at_all, only_owner_edit=only_owner_edit)
        if left_open_ids or left_user_ids:
            add_resp = self.add_chatter_all(chat_id=response.chat_id,
                                            user_ids=left_user_ids, open_ids=left_open_ids)
            response.invalid_open_ids.extend(add_resp.invalid_open_ids)
            response.invalid_user_ids.extend(add_resp.invalid_user_ids)
        return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""消息卡片

https://open.feishu.cn/document/ukTMukTMukTM/uczM3QjL3MzN04yNzcDN
"""
from enum import Enum
from typing import List, Optional, Union, Literal

from pydantic import BaseModel, Field

from .message import SendMsgType


class CardTag(str, Enum):
    # tag for TextObject
    PLAIN_TEXT = "plain_text"
    LARK_MD = "lark_md"

    # tag for CardModule
    DIV = "div"
    HR = "hr"
    IMG = "img"
    ACTION = "action"
    NOTE = "note"

    # tag for CardElement
    BUTTON = "button"
    SELECT_STATIC = "select_static"
    SELECT_PERSON = "select_person"
    OVERFLOW = "overflow"
    DATE_PICKER = "date_picker"
    PICKER_TIME = "picker_time"
    PICKER_DATETIME = "picker_datetime"

    # tag alias for 美学
    TIME_PICKER = PICKER_TIME
    DATETIME_PICKER = PICKER_DATETIME
    PICKER_DATE = DATE_PICKER


class CardTemplate(str, Enum):
    BLUE = "blue"
    WATHET = "wathet"
    TURQUOISE = "turquoise"
    GREEN = "green"
    YELLOW = "yellow"
    ORANGE = "orange"
    RED = "red"
    CARMINE = "carmine"
    VIOLET = "violet"
    PURPLE = "purple"
    INDIGO = "indigo"
    GREY = "grey"


class CardImgMode(str, Enum):
    FIT_HORIZONTAL = "fit_horizontal"
    CROP_CENTER = "crop_center"


class CardLayout(str, Enum):
    BISECTED = "bisected"
    TRISECTION = "trisection"
    FLOW = "flow"


class CardButtonType(str, Enum):
    DEFAULT = "default"
    PRIMARY = "primary"
    DANGER = "danger"


class CardObject(BaseModel):
    """卡片对象

    https://open.feishu.cn/document/ukTMukTMukTM/uUzNwUjL1cDM14SN3ATN
    """


class CardTextObject(CardObject):
    """text对象

    https://open.feishu.cn/document/ukTMukTMukTM/uUzNwUjL1cDM14SN3ATN

    {
        "tag": "plain_text",
        "content": "single line information",
        "lines": 1
    }
    {
        "tag": "plain_text",
        "template":
        "i18n": {
            "zh_cn": "中文文本",
            "en_us": "English text",
            "ja_jp": "日本語文案"
        }
    }
    """
    tag: CardTag = CardTag.PLAIN_TEXT
    content: str
    i18n: Optional[dict] = None
    lines: Optional[int] = None


class CardFieldObject(CardObject):
    """field对象

    https://open.feishu.cn/document/ukTMukTMukTM/uYzNwUjL2cDM14iN3ATN
    {
        "is_short": true,
        "text": {
            "tag": "plain_text",
            "content": "text"
        }
    }
    """
    is_short: bool
    text: CardTextObject


class CardURLObject(CardObject):
    """url对象

    https://open.feishu.cn/document/ukTMukTMukTM/uczNwUjL3cDM14yN3ATN

    {
        "url": "https://www.baidu.com",
        "android_url": "https://developer.android.com/",
        "ios_url": "https://developer.apple.com/",
        "pc_url": "https://www.windows.com"
    }
    """
    url: str
    android_url: str
    ios_url: str
    pc_url: str


class CardOptionObject(CardObject):
    """option对象

    https://open.feishu.cn/document/ukTMukTMukTM/ugzNwUjL4cDM14CO3ATN

    {
        "text": {
            "tag": "plain_text",
            "content": "Option"
        },
        "value": "option"
    }
    """
    text: Optional[CardTextObject] = None
    value: str
    url: Optional[str] = None
    multi_url: Optional[CardURLObject] = None


class CardConfirmObject(CardObject):
    """confirm对象

    https://open.feishu.cn/document/ukTMukTMukTM/ukzNwUjL5cDM14SO3ATN
    {
        "title":{
            "tag":"plain_text",
            "content":"title"
        },
        "text":{
            "tag":"plain_text",
            "content":"content"
        }
    }
    """
    title: CardTextObject
    text: CardTextObject


class CardElement(BaseModel):
    """元素"""
    tag: CardTag


class CardImageElement(CardElement):
    """image元素

    https://open.feishu.cn/document/ukTMukTMukTM/uAzNwUjLwcDM14CM3ATN
    放置在内容模块的extra的图片尺寸固定为64*64
    放置在备注模块的elements的图片尺寸固定为16*16
    {
        "tag": "img",
        "img_key": "f32**************************cf7",
        "alt": {
            "tag": "plain_text",
            "content": "Note image size：16*16"
        }
    }
    """
    tag: Literal[CardTag.IMG] = CardTag.IMG
    img_key: str
    alt: CardTextObject


class CardButtonElement(CardElement):
    """button元素

    https://open.feishu.cn/document/ukTMukTMukTM/uEzNwUjLxcDM14SM3ATN
    Args:
        url 跳转链接，和multi_url互斥
        multi_url 多端跳转链接
        type 样式
        value 交互元素的value字段值仅支持key-value形式的json结构，且key为String类型，示例如下：
                "value":{
                    "key-1":Object-1,
                    "key-2":Object-2,
                    "key-3":Object-3,
                        ······
                }
    整体示例
    {
        "tag":"button",
        "text":{
            "tag":"lark_md",
            "content":"default style"
        },
        "type":"default",
        "value":{
            "key":"value"
        }
    }
    """
    tag: Literal[CardTag.BUTTON] = CardTag.BUTTON
    text: CardTextObject
    url: Optional[str] = None
    multi_url: Optional[CardURLObject] = None
    type: Optional[CardButtonType] = None
    value: Optional[dict] = None
    confirm: Optional[CardConfirmObject] = None


class CardSelectElement(CardElement):
    """selectMenu元素

    https://open.feishu.cn/document/ukTMukTMukTM/uIzNwUjLycDM14iM3ATN
    selectMenu属于交互元素的一种，可用于内容块的extra字段和交互块的elements字段。

    选项模式（"tag":"select_static")：通过options字段配置选项，支持对多个选项进行展示供用户选择。
    选人模式（"tag":"select_person")：通过options字段配置待选人员，无则使用当前群组作为待选人员。

    {
        "tag": "select_static",
        "placeholder": {
            "tag": "plain_text",
            "content": "Option-common mode"
        },
        "value": {
            "key": "value"
        },
        "options": [
            {
                "text": {
                     "tag": "plain_text",
                     "content": "james"
                },
                "value": "james"
            },
            {
                "text": {
                    "tag": "plain_text",
                    "content": "joy"
                },
                "value": "joy"
            },
            {
                "text": {
                    "tag": "plain_text",
                    "content": "james_1"
                },
                "value": "james_1"
            },
            {
                "text": {
                    "tag": "plain_text",
                    "content": "joy_1"
                },
                "value": "joy_1"
            }
        ]
    }
    """
    tag: Literal[CardTag.SELECT_STATIC, CardTag.SELECT_PERSON] = CardTag.SELECT_STATIC
    placeholder: Optional[CardTextObject] = None
    initial_option: Optional[str] = None
    options: Optional[CardOptionObject] = None
    value: Optional[dict] = None
    confirm: Optional[CardConfirmObject] = None


class CardOverflowElement(CardElement):
    """overflow元素

    https://open.feishu.cn/document/ukTMukTMukTM/uMzNwUjLzcDM14yM3ATN
    {
        "tag": "overflow",
        "options": [
            {
                "text": {
                    "tag": "plain_text",
                    "content": "Option-1"
                },
                "value": "option-1"
            },
            {
                "text": {
                    "tag": "plain_text",
                    "content": "Option-2"
                },
                "value": "option-2"
            },
            {
                "text": {
                    "tag": "plain_text",
                    "content": "Option-3”
                },
                "value": "option-3"
            }
        ],
        "value": {
            "key": "value"
        }
    }
    """
    tag: Literal[CardTag.OVERFLOW] = CardTag.OVERFLOW
    options: List[CardOptionObject]
    value: Optional[dict] = None
    confirm: Optional[CardConfirmObject] = None


class CardPickerElement(CardElement):
    """datePicker元素

    https://open.feishu.cn/document/ukTMukTMukTM/uQzNwUjL0cDM14CN3ATN
    Args:
        tag 可以是"date_picker", "picker_time", "picker_datetime"
        initial_date "YYYY-MM-DD"
        initial_time "HH:mm"
        initial_datetime "YYYY-MM-DD HH:mm"
        placeholder 占位符, 无初始值时必填
    {
        "tag": "date_picker",
        "placeholder": {
            "tag": "plain_text",
            "content": "Please select date"
        },
        "value": {
            "key": "value"
        }
    }
    """
    tag: Literal[CardTag.DATE_PICKER] = CardTag.DATE_PICKER
    initial_date: Optional[str] = None
    initial_time: Optional[str] = None
    initial_datetime: Optional[str] = None
    placeholder: Optional[str] = None
    value: Optional[dict] = None
    confirm: Optional[CardConfirmObject] = None


CardGenericElement = Union[
    CardButtonElement, CardPickerElement, CardImageElement, CardSelectElement, CardOverflowElement, dict]


class CardConfig(BaseModel):
    """卡片配置

    https://open.feishu.cn/document/ukTMukTMukTM/uAjNwUjLwYDM14CM2ATN
    """
    wide_screen_mode: Optional[bool] = None
    enable_forward: Optional[bool] = None


class CardHeader(BaseModel):
    """卡片标题

    https://open.feishu.cn/document/ukTMukTMukTM/ukTNwUjL5UDM14SO1ATN
    """
    title: CardTextObject
    template: Optional[CardTemplate] = None


class CardModule(BaseModel):
    """模块"""
    tag: CardTag


class CardDivModule(CardModule):
    """内容模块

    https://open.feishu.cn/document/ukTMukTMukTM/uMjNwUjLzYDM14yM2ATN
    {
        "tag": "div",
        "text": {
            "tag": "plain_text",
            "content": "Content module"
        },
        "fields": [
            {
                "is_short": false,
                "text": {
                    "tag": "lark_md",
                    "content": "**module:**\nContent module（div）"
                }
            },
            {
                "is_short": false,
                "text": {
                    "tag": "lark_md",
                    "content": "**function:**\nNew function"
                }
            }
        ],
        "extra": {
            "tag": "img",
            "img_key": "f32******************5cf7",
            "alt": {
                "tag": "plain_text",
                "content": "alt_content"
            }
        }
    }
    """
    tag: Literal[CardTag.DIV] = CardTag.DIV
    text: CardTextObject
    fields_: List[CardFieldObject] = Field([], alias="fields")
    extra: Optional[CardGenericElement] = None


class CardHrModule(CardModule):
    """分割线模块
    {
        "tag": "hr",
    }
    """
    tag: Literal[CardTag.HR] = CardTag.HR


class CardImgModule(CardModule):
    """图片模块

    https://open.feishu.cn/document/ukTMukTMukTM/uUjNwUjL1YDM14SN2ATN
    {
        "tag":"img",
        "title":{
           "tag":"plain_text",
           "content":"Block-img"
        },
        "img_key":"9f6********************315",
        "alt":{
           "tag":"plain_text",
           "content":"Block-img"
        }
    }
    """
    tag: Literal[CardTag.IMG] = CardTag.IMG
    img_key: str
    alt: CardTextObject
    title: Optional[CardTextObject] = None
    mode: Optional[CardImgMode] = None


class CardActionModule(CardModule):
    """交互模块"""
    tag: Literal[CardTag.ACTION] = CardTag.ACTION
    actions: List[CardGenericElement]
    layout: Optional[CardLayout] = None


CardGenericModule = Union[CardHrModule, CardImgModule, CardActionModule, CardDivModule, dict]


class I18nContent(BaseModel):
    en_us: Optional[List[CardGenericModule]] = None
    zh_cn: Optional[List[CardGenericModule]] = None


class CardContent(BaseModel):
    config: Optional[CardConfig] = None
    header: Optional[CardHeader] = None
    card_link: Optional[CardURLObject] = None
    elements: Optional[List[CardGenericModule]] = None
    i18n_elements: Optional[I18nContent] = None


class CardMessage(BaseModel):
    """卡片消息

    https://open.feishu.cn/document/ukTMukTMukTM/uYTNwUjL2UDM14iN1ATN
    Args:
        chat_id 同下
        open_id 同下
        user_id 同下
        email 如下
            - 私聊的话open_id/user_id/email三个里面要填一个
            - 群聊的话必须填chat_id
            - 如果都提供的话，优先级为 chat_id > open_id > user_id > email
        root_id 回复消息时，对应消息的消息ID，选填
        msg_type 消息类型, = "interactive"
        card 卡片内容, 见CardContent
        update_multi 是否所有用户共享同一张卡片, 默认为False

    示例
    {
        "chat_id": "oc_abcdefg1234567890",
        "msg_type": "interactive",
        "root_id":"om_4*********************ad8",
        "update_multi":false,
        "card": {
                // card content
        }
    }
    """
    chat_id: Optional[str] = None
    open_id: Optional[str] = None
    user_id: Optional[str] = None
    email: Optional[str] = None
    root_id: Optional[str] = None
    msg_type: str = SendMsgType.INTERACTIVE
    card: CardContent
    update_multi: Optional[bool] = None



import hashlib
import base64
from Crypto.Cipher import AES


class AESCipher(object):
    def __init__(self, key):
        self.bs = AES.block_size
        self.key = hashlib.sha256(AESCipher.str_to_bytes(key)).digest()

    @staticmethod
    def str_to_bytes(data):
        u_type = type(b"".decode('utf8'))
        if isinstance(data, u_type):
            return data.encode('utf8')
        return data

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[len(s) - 1:])]

    def decrypt(self, enc):
        iv = enc[:AES.block_size]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(enc[AES.block_size:]))

    def decrypt_string(self, enc):
        enc = base64.b64decode(enc)
        return self.decrypt(enc).decode('utf8')


from .auth import AuthAPI
from .base import allow_async_call
from .bot import BotAPI
from .card import CardAPI
from .message import MessageAPI
from .cloud import CloudAPI
from .message2 import MessageAPI2


class FeishuAPI(AuthAPI, CardAPI, BotAPI, CloudAPI, MessageAPI2):
    """飞书API

    实现API的简单示例::

    >>> class DemoAPI(BaseAPI):
    ...     @allow_async_call
    ...     def demo_api(self, arg1, arg2) -> dict:
    ...         params, payload = do_something_with_arguments(arg1, arg2)
    ...         result = self.client.request(api, params, payload)
    ...         result = do_something_with_result(result)
    ...         return result

    这里client就是FeishuClient的示例，封装了统一调度的网络方法request
    api = /auth/v3/tenant_access_token/internal/
    params, payload为url参数和body参数，可为空不传

    注意:
        为了让allow_async_call的hack起作用，
        请一定要通过`result = self.client.request`或者`result = self.client.fetch_url`来发起通用请求，
        否则可能会造在异步代码中执行同步请求，卡住event_loop的问题
    """

    def __init__(self, feishu_client: "FeishuClient"):
        self.client = feishu_client
        # 执行allow_async_call里面的hack
        self.dummy()

    @allow_async_call
    def dummy(self):
        """这是一个用来初始化async函数的API"""
        pass



# /usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Tuple, Dict

from .base import BaseAPI, allow_async_call


class AuthAPI(BaseAPI):

    @allow_async_call
    def get_tenant_access_token(self) -> Tuple[str, int]:
        """获取自建应用的token

        https://open.feishu.cn/document/ukTMukTMukTM/uIjNz4iM2MjLyYzM

        Returns:
            Tuple[token, expire]
        """
        api = "/auth/v3/tenant_access_token/internal/"
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
        )
        result = self.client.request("POST", api=api, payload=payload, auth=False)
        return result["tenant_access_token"], result["expire"]

    def resend_app_ticket(self):
        """重新推送app_ticket

        https://open.feishu.cn/document/ukTMukTMukTM/uQjNz4CN2MjL0YzM
        """
        NotImplementedError

    @allow_async_call
    def get_user_info(self, user_id: str,
                      user_id_type: str = 'open_id',
                      department_id_type: str = 'open_department_id') -> Dict:
        """获取单个用户信息（user_id_type和department_id_type设置默认值）

        https://open.feishu.cn/open-apis/contact/v3/users/:user_id

        Returns:
            Dict
        """
        api = "/contact/v3/users/" + user_id
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
        )
        params = dict(
            user_id_type=user_id_type,
            department_id_type=department_id_type,
        )
        result = self.client.request("GET", api=api, payload=payload, params=params)
        return result

    @allow_async_call
    def get_user_by_department(self, department_id: str, user_id_type: str = 'open_id',
                               department_id_type: str = 'open_department_id'):
        """获取部门直属用户列表

        https://open.feishu.cn/open-apis/contact/v3/users/find_by_department

        Returns:
            json
        """
        api = "/contact/v3/users/find_by_department"
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
        )
        params = dict(
            user_id_type=user_id_type,
            department_id_type=department_id_type,
            department_id=department_id,
        )
        result = self.client.request("GET", api=api, payload=payload, params=params)
        return result

    @allow_async_call
    def get_department_by_id(self, department_id: str, user_id_type: str = 'open_id',
                             department_id_type: str = 'open_department_id'):
        """获取单个部门信息

        https://open.feishu.cn/open-apis/contact/v3/departments/:department_id

        Returns:
            json
        """
        api = "/contact/v3/departments/" + department_id
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
        )
        params = dict(
            user_id_type=user_id_type,
            department_id_type=department_id_type,
        )
        result = self.client.request("GET", api=api, payload=payload, params=params)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop
from typing import Union


class FeishuBaseClient(ABC):
    logger = logging.getLogger("feishu")

    app_id: str
    app_secret: str
    run_async: bool
    event_loop: AbstractEventLoop

    @abstractmethod
    def request(self, method: str, api: str, params: dict = {}, payload: dict = {},
                data: dict = {}, files: dict = {}, auth: str = True) -> Union[dict, bytes]:
        """发起请求

        Args:
            method: "GET" or "POST"
            api: 对应功能的API Path, e.g. "/user/v1/union_id/batch_get/list"
            params: HTTP的URL参数
            payload: Body的参数, 会序列化为json
            data: form-data
            files: multipart/form-data
            auth: 是否需要验证, 只有token类API需要设为False

        Returns:
            一个解析好的返回dict，为飞书的标准格式
            code: 0为正常
            msg: 出错信息
            data: 真正的数据信息

        Raises:
            FeishuException
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""飞书通讯录事件

https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/contact-v3/department/field-overview
"""
from .base import EventContent


class UserCreatedEvent(EventContent):
    """
    事件类型：员工入职
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/contact-v3/user/events/created
    """

    @staticmethod
    def event_type():
        return "contact.user.created_v3"


class UserUpdatedEvent(EventContent):
    """
    事件类型：员工信息被修改
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/contact-v3/user/events/updated
    """

    @staticmethod
    def event_type():
        return "contact.user.updated_v3"


class UserDeletedEvent(EventContent):
    """
    事件类型：员工离职
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/contact-v3/user/events/deleted
    """

    @staticmethod
    def event_type():
        return "contact.user.deleted_v3"


class DepartmentCreatedEvent(EventContent):
    """
    事件类型：部门被创建
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/contact-v3/department/events/created
    """

    @staticmethod
    def event_type():
        return "contact.department.created_v3"


class DepartmentUpdatedEvent(EventContent):
    """
    事件类型：员工信息被修改
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/contact-v3/department/events/updated
    """

    @staticmethod
    def event_type():
        return "contact.department.updated_v3"


class DepartmentDeletedEvent(EventContent):
    """
    事件类型：部门信息被修改
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/contact-v3/department/events/deleted
    """

    @staticmethod
    def event_type():
        return "contact.department.deleted_v3"



#!/usr/bin/env python3.8

import json
import typing as t

from django_feishu_sdk.decrypt import AESCipher
from django_feishu_sdk.settings import api_settings
from django_feishu_sdk.models.events import *
from django_feishu_sdk.utils import InvalidEventException


class EventManager(object):
    event_callback_map = dict()
    event_type_map = dict()
    _event_list = [MessageReceiveEvent, UrlVerificationEvent, MessageReadEvent]

    def __init__(self):
        for event in EventManager._event_list:
            EventManager.event_type_map[event.event_type()] = event

    def register(self, event_type: str) -> t.Callable:
        def decorator(f: t.Callable) -> t.Callable:
            self.register_handler_with_event_type(event_type=event_type, handler=f)
            return f

        return decorator

    @staticmethod
    def register_handler_with_event_type(event_type, handler):
        EventManager.event_callback_map[event_type] = handler

    @staticmethod
    def get_handler_with_event(request):
        token = api_settings.FEISHU_VERIFY_TOKEN
        encrypt_key = api_settings.FEISHU_ENCRYPT_KEY
        dict_data = json.loads(request.body)
        # dict_data = decrypt_aes(encrypt_key, dict_data)
        dict_data = EventManager._decrypt_data(encrypt_key, dict_data)
        print('事件解密后dict_data:', dict_data)
        callback_type = dict_data.get("type")
        # only verification data has callback_type, else is event
        if callback_type == "url_verification":
            if dict_data.get('token') != token:
                raise Exception("VERIFICATION_TOKEN is invalid")
            event = UrlVerificationEvent(dict_data)
            return EventManager.event_callback_map.get(event.event_type()), event

        # only handle event v2
        schema = dict_data.get("schema")
        if schema is None:
            raise InvalidEventException("request is not callback event(v2)")

        # get event_type
        event_type = dict_data.get("header").get("event_type")
        # build event
        event = EventManager.event_type_map.get(event_type)(request, dict_data, token, encrypt_key)
        # get handler
        print(EventManager.event_callback_map.get(event_type))
        print(event)
        return EventManager.event_callback_map.get(event_type), event

    @staticmethod
    def _decrypt_data(encrypt_key, data):
        encrypt_data = data.get("encrypt")
        print('encrypt_data:', encrypt_data)
        if encrypt_key == "" and encrypt_data is None:
            # data haven't been encrypted
            return data
        if encrypt_key == "":
            raise Exception("ENCRYPT_KEY is necessary")
        cipher = AESCipher(encrypt_key)

        return json.loads(cipher.decrypt_string(encrypt_data))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""飞书消息和群组事件

https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/im-v1/introduction
"""
from .base import EventContent
from ...utils import Dict2Obj


class UrlVerificationEvent(EventContent):
    """
    事件类型：应用首次验证——请求地址配置
    飞书文档地址：https://open.feishu.cn/apps/cli_a27a016dd6fbd013/event
    """

    def __init__(self, dict_data):
        # event check and init
        self.event = Dict2Obj(dict_data)

    @staticmethod
    def event_type():
        return "url_verification"


class MessageReceiveEvent(EventContent):
    """
    事件类型：接收消息
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/im-v1/message/events/receive
    """

    @staticmethod
    def event_type():
        return "im.message.receive_v1"


class MessageReadEvent(EventContent):
    """
    事件类型：已读消息
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/im-v1/message/events/message_read
    """

    @staticmethod
    def event_type():
        return "im.message.message_read_v1"


class AddBotEvent(EventContent):
    """
    事件类型：机器人进群
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/im-v1/chat-member-bot/events/added
    """

    @staticmethod
    def event_type():
        return "im.chat.member.bot.added_v1"


class RemoveBotEvent(EventContent):
    """
    事件类型：机器人被移出群
    飞书文档地址：https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/im-v1/chat-member-bot/events/deleted
    """

    @staticmethod
    def event_type():
        return "im.chat.member.bot.deleted_v1"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""消息

https://open.feishu.cn/document/ukTMukTMukTM/ucDO1EjL3gTNx4yN4UTM
"""
from .card import *
from .message import (Message, TextMessage, ImageMessage, PostMessage, ShareChatMessage,
                      Content, PostContent, ImageContent, ShareChatContent, TextContent,
                      SendMsgType, ReadUser, PostTag, PostElement, Post, PostAElement, PostAtElement, PostImgElement,
                      PostTextElement, I18nPost, BatchMessage, BatchSendResponse)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""消息管理相关API

包括:
- 发送文本、图片、富文本、名片消息
- 发送互动卡片消息
- 图片文本上传
- 发送通知消息

注意，被动消息接收在event.py中
"""
from enum import Enum
from typing import Optional, Union, Type, List

from .base import BaseAPI, allow_async_call
from ..settings import api_settings
from ..errors import ERRORS, FeishuError
from ..models import (Message, TextMessage, TextContent, SendMsgType, Content, ImageMessage, PostMessage,
                      ShareChatMessage, ImageContent, I18nPost, PostContent, ShareChatContent, BatchSendResponse,
                      BatchMessage)

fileobj = Type
Image = Type


class ImageType(str, Enum):
    MESSAGE = "message"
    AVATAR = "avatar"

    def __str__(self):
        return self.value


def create_message(msg_type: SendMsgType, content: Content,
                   chat_id: str, open_id: str, user_id: str, email: str, root_id: str) -> Message:
    message_cls = {
        SendMsgType.TEXT: TextMessage,
        SendMsgType.IMAGE: ImageMessage,
        SendMsgType.POST: PostMessage,
        SendMsgType.SHARE_CHAT: ShareChatMessage,
    }[msg_type]
    msg = message_cls(
        msg_type=msg_type,
        content=content
    )
    if root_id:
        msg.root_id = root_id

    if chat_id:
        msg.chat_id = chat_id
    elif open_id:
        msg.open_id = open_id
    elif user_id:
        msg.user_id = user_id
    elif email:
        msg.email = email

    return msg


class MessageAPI(BaseAPI):
    """消息管理相关API

    https://open.feishu.cn/document/ukTMukTMukTM/uUjNz4SN2MjL1YzM
    """

    @allow_async_call
    def send(self, message: Union[Message, dict]) -> Optional[str]:
        """发送/v4/send请求, 返回message_id

        Args:
            message: Message类型或者一个简单的dict

        Returns:
            str: message_id

        Usages::

        >>> # from feishu import Message, SendMsgType, TextContent
        >>> # from feishu import FeishuClient
        >>> msg = Message(open_id='xxx', msg_type=SendMsgType.Text, content=TextContent(text="abc"))
        >>> # or
        >>> # msg = {"open_id": "xxx", "msg_type": "text", "content": {"text": "abc"}}
        >>>
        >>> client = FeishuClient(...)
        >>> client.send(msg)
        """
        api = "/message/v4/send/"
        if isinstance(message, Message):
            payload = message.dict(exclude_none=True)
        else:
            payload = message
        result = self.client.request("POST", api=api, payload=payload)
        return result.get("data", {}).get("message_id")

    @allow_async_call
    def send_text(self, text: str, open_id: str = '', user_id: str = '', email: str = '',
                  chat_id: str = '', root_id: str = '') -> Optional[str]:
        """发送文本消息

        Args:
            text: 待发送的文本信息
            open_id: 用户的飞信ID(自建应用)
            user_id: 用户的应用ID(三方应用)
            email: 用户的邮箱
            chat_id: 群ID, 以上4种ID必须提供一种, 优先级为chat_id>open_id>user_id>email
            root_id: 回复消息所对应的呃消息id, 可选
        """
        msg = create_message(SendMsgType.TEXT, content=TextContent(text=text),
                             root_id=root_id, chat_id=chat_id, open_id=open_id, user_id=user_id, email=email)

        if text.strip():
            return self.send(msg)
        else:
            self.logger.warning(f"text为空, 文本消息未发送: msg={msg}")

    @allow_async_call
    def send_image(self, image: Union[bytes, "fileobj", "Image"] = '',
                   image_file: str = '', image_url: str = '', image_key: str = '',
                   open_id: str = '', user_id: str = '', email: str = '',
                   chat_id: str = '', root_id: str = '') -> Optional[str]:
        """发送文本消息

        Args:
            image: 待发送的图片，可以是PIL.Image, bytes, fileobj
            image_file: 待发送图片文件路径
            image_url: 待发送图片的url
            image_key: 待发送图片的key, 飞书专用，以上4个必须提供一种，优先级为: image_key>image>image_file>image_url
            open_id: 用户的飞信ID(自建应用)
            user_id: 用户的应用ID(三方应用)
            email: 用户的邮箱
            chat_id: 群ID, 以上4种ID必须提供一种, 优先级为chat_id>open_id>user_id>email
            root_id: 回复消息所对应的呃消息id, 可选
        """
        if not image_key:
            image_key = self._get_image_key(image=image, image_file=image_file, image_url=image_url)

        print("image_key2", image_key)
        if image_key:
            msg = create_message(SendMsgType.IMAGE, content=ImageContent(image_key=image_key),
                                 root_id=root_id, chat_id=chat_id, open_id=open_id, user_id=user_id, email=email)
            print("msg=", msg.dict(exclude_none=True))
            return self.send(msg)
        else:
            self.logger.warning(f"没有提供image_key, image_url, image_file, 或image，图片未发送: msg={msg}")

    @allow_async_call
    def send_post(self, post: Union[dict, I18nPost],
                  open_id: str = '', user_id: str = '', email: str = '',
                  chat_id: str = '', root_id: str = '') -> Optional[str]:
        """发送富文本消息

        Args:
            post: dict或者I18nPost类型
            open_id: 用户的飞信ID(自建应用)
            user_id: 用户的应用ID(三方应用)
            email: 用户的邮箱
            chat_id: 群ID, 以上4种ID必须提供一种, 优先级为chat_id>open_id>user_id>email
            root_id: 回复消息所对应的呃消息id, 可选

        post参数示例
        {
            "zh_cn": {
                "title": "我是一个标题",
                "content": [
                    [
                        {
                            "tag": "text",
                            "un_escape": true,
                            "text": "第一行&nbsp;:"
                        },
                        {
                            "tag": "a",
                            "text": "超链接",
                            "href": "http://www.feishu.cn"
                        },
                        {
                            "tag": "at",
                            "user_id": "ou_18eac85d35a26f989317ad4f02e8bbbb"
                        }
                    ],
                ]
            }
        }
        """
        msg = create_message(msg_type=SendMsgType.POST, content=PostContent(post=post),
                             root_id=root_id, chat_id=chat_id, open_id=open_id, user_id=user_id, email=email)
        if not post.get('zh_cn') and not post.get('en_us'):
            self.logger.warning(f"没有提供zh_cn/en_us内容, 富文本未发送: msg={msg}")
        else:
            return self.send(msg)

    @allow_async_call
    def send_share_chat(self, share_chat_id: str,
                        open_id: str = '', user_id: str = '', email: str = '',
                        chat_id: str = '', root_id: str = '') -> Optional[str]:
        """发送群名片

         Args:
            share_chat_id: 群名片ID
            open_id: 用户的飞信ID(自建应用)
            user_id: 用户的应用ID(三方应用)
            email: 用户的邮箱
            chat_id: 群ID, 以上4种ID必须提供一种, 优先级为chat_id>open_id>user_id>email
            root_id: 回复消息所对应的呃消息id, 可选
        """
        msg = create_message(msg_type=SendMsgType.SHARE_CHAT, content=ShareChatContent(share_chat_id=share_chat_id),
                             root_id=root_id, chat_id=chat_id, open_id=open_id, user_id=user_id, email=email)
        if not share_chat_id.strip():
            return self.send(msg)
        else:
            self.logger.warning(f"share_chat_id为空, 群名片未分享: msg={msg}")

    @allow_async_call
    def batch_send_all(self, message: Union[Message, dict], department_ids: List[str], open_ids: List[str],
                       user_ids: List[str]) -> BatchSendResponse:
        """批量发送消息

        Args:
            message: 消息可以是文本/图片/富文本/群名片/卡片
            department_ids: API有200个的限制，这里取消了200个的限制，改为多次发送，并合并结果
            open_ids: API有200个的限制，这里取消了200个的限制，改为多次发送，并合并结果
            user_ids: API有200个的限制，这里取消了200个的限制，改为多次发送，并合并结果
        """
        slice_size = api_settings.FEISHU_BATCH_SEND_SIZE
        response = BatchSendResponse()
        while department_ids or open_ids or user_ids:
            resp = self.batch_send(message, department_ids=department_ids, open_ids=open_ids, user_ids=user_ids)
            response.message_id = resp.message_id
            response.message_ids.append(resp.message_id)
            response.invalid_department_ids.extend(resp.invalid_department_ids)
            response.invalid_open_ids.extend(resp.invalid_open_ids)
            response.invalid_user_ids.extend(resp.invalid_user_ids)

            department_ids = department_ids[slice_size:]
            open_ids = open_ids[slice_size:]
            user_ids = user_ids[slice_size:]

        return response

    @allow_async_call
    def batch_send(self, message: Union[Message, dict], department_ids: List[str], open_ids: List[str],
                   user_ids: List[str]) -> BatchSendResponse:
        """和batch_send_all的区别是id有200个的限制"""
        api = "/message/v4/batch_send/"
        batch_msg = BatchMessage(
            department_ids=department_ids,
            open_ids=open_ids,
            user_ids=user_ids,
            msg_type=message.msg_type,
            content=message.content,
        )
        result = self.client.request(method="POST", api=api, payload=batch_msg.dict())
        return BatchSendResponse(**(result.get("data") or {}))

    @allow_async_call
    def upload_image(self, image: Union[bytes, "fileobj"], image_type: ImageType = "message") -> Optional[str]:
        """上传图片

        Args:
            image: 可以是bytes或fileobj
            image_type: 图片类型, 可以是message/avatar

        Returns:
            str: image_key
        """
        api = "/image/v4/put/"
        data = {"image_type": image_type}
        files = {"image": image}
        result = self.client.request(method="POST", api=api, data=data, files=files)
        return result.get("data", {}).get("image_key")

    @allow_async_call
    def get_image(self, image_key: str) -> bytes:
        """获取图片数据"""
        api = "/image/v4/get"
        params = {"image_key": image_key}
        # TODO: 这个API要搞清楚是不是直接返回，是的话需要修改通用的request类
        return self.client.request(method="GET", api=api, params=params)

    @allow_async_call
    def _get_image_key(self, image: Union[bytes, "fileobj", "Image"] = '',
                       image_file: str = '', image_url: str = '') -> str:
        """获取image_key

        Args:
            image: 图片本身
            image_file: 图片文件的路径
            image_url: 图片链接

        Returns:
            image_key, 如果失败则raise FeishuError
        """
        image_key = ''
        if image:
            try:
                if hasattr(image, "tobytes"):
                    image = image.tobytes()

                image_key = self.upload_image(image)
                assert image_key
            except Exception as e:
                raise FeishuError(ERRORS.INVALID_IMAGE_FILE_OR_CONTENT,
                                  f"上传图片失败: {str(e)} image={image[:20]}...")
        elif image_file:
            try:
                image_key = self.upload_image(open(image_file, "rb"))
            except Exception as e:
                raise FeishuError(ERRORS.INVALID_IMAGE_FILE_OR_CONTENT,
                                  f"上传图片失败: {str(e)} image_file={image_file}")
        elif image_url:
            try:
                content = self.client.fetch(method="GET", url=image_url)
            except Exception as e:
                raise FeishuError(ERRORS.INVALID_IMAGE_FILE_OR_CONTENT,
                                  f"下载图片失败: {str(e)} image_url={image_url}")
            try:
                image_key = self.upload_image(content)
                print("image_key", image_key)
            except Exception as e:
                raise FeishuError(ERRORS.INVALID_IMAGE_FILE_OR_CONTENT,
                                  f"上传图片失败: {str(e)} content={content[:20]}...")

        return image_key



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""消息卡片相关

- 发送卡片消息，临时卡片消息，删除临时卡片消息
- flask和sanic的blueprint
"""
import logging
from typing import Union, Callable, Optional, Awaitable

from .base import BaseAPI, allow_async_call, decrypt_aes
from ..models import CardMessage, CardContent, SendMsgType

logger = logging.getLogger("feishu")


class CardAPI(BaseAPI):
    @allow_async_call
    def send_card(
            self,
            card: Union[dict, CardMessage],
            update_multi: bool = None,
            open_id: str = "",
            user_id: str = "",
            email: str = "",
            chat_id: str = "",
            root_id: str = "",
    ) -> Optional[str]:
        """发送卡片消息

        https://open.feishu.cn/document/ukTMukTMukTM/uYTNwUjL2UDM14iN1ATN

        Args:
            card: dict或CardMessage类型
            update_multi: 控制卡片是否是共享卡片, 默认位False
            open_id: 用户的飞信ID(自建应用)
            user_id: 用户的应用ID(三方应用)
            email: 用户的邮箱
            chat_id: 群ID, 以上4种ID必须提供一种, 优先级为chat_id>open_id>user_id>email
            root_id: 回复消息所对应的呃消息id, 可选
        Returns:
            message_id

        请求body示例
        {
           "chat_id": "oc_abcdefg1234567890",
           "msg_type": "interactive",
           "root_id":"om_4*********************ad8",
           "update_multi":false,
           "card": {
                // card content
            }
        }

        card参数示例
        {
            "config": {
                "wide_screen_mode": true
            },
            "header": {
                "title": {
                    "tag": "plain_text",
                    "content": "this is header"
                }
            },
            "elements": [
                {
                    "tag": "div",
                    "text": {
                        "tag": "plain_text",
                        "content": "This is a very very very very very very very long text;"
                    }
                },
                {
                    "tag": "action",
                    "actions": [
                        {
                            "tag": "button",
                            "text": {
                                "tag": "plain_text",
                                "content": "Read"
                            },
                            "type": "default"
                        }
                    ]
                }
            ]
        }
        """
        api = "/message/v4/send/"

        msg = CardMessage(
            msg_type=SendMsgType.INTERACTIVE, card=card, update_multi=update_multi
        )
        if root_id:
            msg.root_id = root_id

        if chat_id:
            msg.chat_id = chat_id
        elif open_id:
            msg.open_id = open_id
        elif user_id:
            msg.user_id = user_id
        elif email:
            msg.email = email

        payload = msg.dict(exclude_none=True)
        result = self.client.request("POST", api=api, payload=payload)
        return result.get("data", {}).get("message_id")

    def send_ephemeral_card(self, card: Union[dict, CardMessage]) -> str:
        """发送临时卡片消息

        https://open.feishu.cn/document/ukTMukTMukTM/uETOyYjLxkjM24SM5IjN

        使用场景
            临时消息卡片多用于群聊中用户与机器人交互的中间态。
            例如在群聊中用户需要使用待办事项类bot创建一条提醒，
            bot 发送了可设置提醒日期和提醒内容的一张可交互的消息卡片，
            此卡片在没有设置为临时卡片的情况下为群内全员可见，既群内可看见该用户与 bot 交互的过程。
            而设置为临时卡片后，交互过程仅该用户可见，群内其他成员只会看到最终设置完成的提醒卡片。

        临时消息卡片可降低群消息的信噪比，并间接增加 bot 通知的用户触达。

        注意, card对象必须提供chat_id
        """
        api = "/ephemeral/v1/send"

        if isinstance(card, CardMessage):
            payload = card.dict(exclude_unset=True)
        else:
            payload = dict(card)

        result = self.client.request("POST", api=api, payload=payload)
        return result.get("data", {}).get("message_id")

    def delete_ephemeral_card(self, message_id: str):
        """删除临时卡片

        失败会raise FeishuError, (code和msg对应飞书平台的code/msg)
        """
        api = "/ephemeral/v1/delete"
        payload = {"message_id": message_id}
        self.client.request("POST", api=api, payload=payload)



# /usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Tuple

from .base import BaseAPI, allow_async_call


class CloudAPI(BaseAPI):

    @allow_async_call
    def get_cloud_allfiles(self, folder_token):
        """获取文件夹下的清单

        https://open.feishu.cn/open-apis/drive/v1/files

        Returns:
            json
        """
        api = "/drive/v1/files"
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
        )
        params = dict(
            page_size=10,
            folder_token=folder_token,
        )
        result = self.client.request("GET", api=api, payload=payload, params=params)
        # 返回 Body ：
        # {
        #     "code":0,
        #     "msg":"ok",
        #     "tenant_access_token":"xxxxx",
        #     "expire":7200  // 过期时间，单位为秒（两小时失效）
        # }
        return result

    @allow_async_call
    def get_cloud_folderdata(self, folderToken: str):
        """获取文件夹元信息

        https://open.feishu.cn/open-apis/drive/explorer/v2/folder/:folderToken/meta

        Returns:
            json
        """
        api = "/drive/explorer/v2/folder/" + folderToken + "/meta"
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
        )
        params = dict(
        )
        result = self.client.request("GET", api=api, payload=payload, params=params)
        # 返回 Body ：
        # {
        #     "code":0,
        #     "msg":"ok",
        #     "tenant_access_token":"xxxxx",
        #     "expire":7200  // 过期时间，单位为秒（两小时失效）
        # }
        return result

    @allow_async_call
    def creat_cloud_folder(self, name, folder_token: str = ''):
        """在云文件夹下创建文件夹

        https://open.feishu.cn/open-apis/drive/v1/files/create_folder

        Returns:
            json
        """
        api = "/drive/v1/files/create_folder"
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
            name=name,
            folder_token=folder_token,
        )
        result = self.client.request("POST", api=api, payload=payload)
        # 返回 Body ：
        # {
        #     "code":0,
        #     "msg":"ok",
        #     "tenant_access_token":"xxxxx",
        #     "expire":7200  // 过期时间，单位为秒（两小时失效）
        # }
        return result

    @allow_async_call
    def creat_cloud_table(self, title, folder_token: str = ''):
        """在云文件夹下创建表格
        https://open.feishu.cn/open-apis/sheets/v3/spreadsheets

        Args:
            title: 表格名字
            folder_token: 表格所在文件
        Returns:
            json
        Raises:
            FeishuException
        """
        api = "/sheets/v3/spreadsheets"
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
            title=title,
            folder_token=folder_token,
        )
        result = self.client.request("POST", api=api, payload=payload)
        # 返回 Body ：
        # {
        #     "code":0,
        #     "msg":"ok",
        #     "tenant_access_token":"xxxxx",
        #     "expire":7200  // 过期时间，单位为秒（两小时失效）
        # }
        return result

    @allow_async_call
    def get_cloud_tabledata(self, spreadsheetToken: str):
        """获取表格元数据

        https://open.feishu.cn/open-apis/sheets/v2/spreadsheets/:spreadsheetToken/metainfo

        Returns:
            json
        """
        api = "/sheets/v2/spreadsheets/" + spreadsheetToken + "/metainfo"
        payload = dict(
            app_id=self.client.app_id,
            app_secret=self.client.app_secret,
        )
        result = self.client.request("GET", api=api, payload=payload)
        # 返回 Body ：
        # {
        #     "code":0,
        #     "msg":"ok",
        #     "tenant_access_token":"xxxxx",
        #     "expire":7200  // 过期时间，单位为秒（两小时失效）
        # }
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import asyncio
import logging
import os
import secrets
from asyncio import Future, AbstractEventLoop
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Union, Tuple

import aiohttp
import requests

from .apis import FeishuAPI, _get_or_create_event_loop
from .baseclient import FeishuBaseClient
from .settings import AppType, api_settings
from .errors import FeishuError, ERRORS
from .stores import TokenStore, MemoryStore

logger = logging.getLogger("feishu")


class FeishuClient(FeishuBaseClient, FeishuAPI):
    """飞书开放平台客户端"""

    def __init__(self, app_id: Optional[str] = None, app_secret: Optional[str] = None,
                 app_type: AppType = AppType.TENANT,
                 run_async: bool = False,
                 event_loop: Optional[AbstractEventLoop] = None,
                 endpoint: str = "https://open.feishu.cn/open-apis/",
                 timeout: float = 5,
                 token_store: Optional[TokenStore] = None):
        """初始化

        Args:
            app_id: 飞书自建应用的app_id, 默认则取环境变量中的FEISHU_APP_ID
            app_secret: 飞书自建应用的app_secret, 默认则取环境变量中的FEISHU_APP_SECRET
            app_type: "tenant": 用户自建应用/"user": 第三方应用
            run_async: 是否异步模式, 异步模式下所有外部调用都返回一个asyncio.Future, 默认为False
            event_loop: 若run_async=True, 可以提供event_loop作为async方法的loop
                如果不提供的话，请确保在执行API请求前设置默认loop: asyncio.set_event_loop(loop)
            timeout: 连接超时，其中timeout/3为连接超时，timeout*2/3为读取超时
            endpoint: 飞书平台的endpoint, 一般默认就好
            token_store: 飞书的access_token会在2小时后过期，这里
        """
        allowed_types = AppType.__dict__["_value2member_map_"]
        if app_type not in allowed_types or app_type == "user":
            raise NotImplementedError(f"不支持app_type={app_type}, 暂时支持了tenant自建应用")

        self.app_id = app_id
        self.app_secret = app_secret
        self.app_type = app_type
        self.run_async = run_async
        self.event_loop: Optional[asyncio.AbstractEventLoop] = event_loop
        self.endpoint = endpoint
        self.timeout = timeout

        if not self.app_id:
            # print('FEISHU_APP_ID', os.environ.get(FEISHU_APP_ID, ""))
            self.app_id = api_settings.FEISHU_APP_ID
        if not self.app_secret:
            self.app_secret = api_settings.FEISHU_APP_SECRET

        assert self.app_id and self.app_secret, "必须提供app_id/app_secret，请查看settings配置是否正确"

        while self.endpoint.endswith("/"):
            self.endpoint = self.endpoint[:-1]

        # 内部使用时用api作为namespace更清晰些
        self.api = FeishuAPI(self)
        # 但同时因为继承了FeishuAPI, 所以必须为那些方法提供一个client
        self.client = self

        if self.run_async:
            self.event_loop = event_loop  # lazy initialize in self.request/self.fetch
            self.session_async = None  # lazy initialize in self.request/self.fetch
            self.executor = ThreadPoolExecutor(2)
        else:
            self.session = requests.Session()
            self.executor = None
        self.closed = False
        if not token_store:
            token_store = MemoryStore()
        self.token_store = token_store

    def get_token(self) -> Union[str, Future]:
        if self.run_async:
            async def _get_token_async():
                token_ = await self.event_loop.run_in_executor(self.executor, self.token_store.get, "token")
                if not token_:
                    token_, expire_ = await self.api.get_tenant_access_token()
                    await self.event_loop.run_in_executor(self.executor, self.token_store.set,
                                                          "token", token_, expire_)
                return token_

            return asyncio.ensure_future(_get_token_async(), loop=self.event_loop)
        else:
            token = self.token_store.get("token")
            if not token:
                if self.app_type == AppType.TENANT:
                    token, expire = self.api.get_tenant_access_token()
                    self.token_store.set("token", token, expire)
                else:
                    raise NotImplementedError

            return token

    def request(self,
                method: str,
                api: str,
                params: dict = {},
                payload: dict = {},
                data: dict = {},
                files: dict = {},
                auth: str = True) -> Union[dict, bytes, Future]:
        """发起请求

        Args:
            method: "GET" or "POST"
            api: 对应功能的API Path, e.g. "/user/v1/union_id/batch_get/list"
            params: HTTP的URL参数
            payload: Body的参数, 会序列化为json
            data: Form-Data格式的参数
            files: Multipart-encoded格式的文件参数
            auth: 是否需要验证, 只有token类API需要设为False

        Returns:
            一个解析好的返回dict，为飞书的标准格式
            code: 0为正常
            msg: 出错信息
            data: 真正的数据信息

        Raises:
            FeishuException
        """
        if self.closed:
            raise FeishuError(ERRORS.CLIENT_CLOSED, "client对象已被关闭")

        headers = {
            "Content-Type": "application/json",
        }

        url = self.endpoint + api
        timeout_pair = (self.timeout / 3, self.timeout * 2 / 3)
        if files:
            headers.pop("Content-Type")

        if self.run_async:
            async def do_request_async():
                if auth:
                    token = await self.get_token()
                    headers['Authorization'] = f"Bearer {token}"
                return await self._async_request(
                    method=method, url=url, timeout_pair=timeout_pair, headers=headers,
                    params=params, payload=payload, data=data, files=files)

            future = asyncio.ensure_future(
                do_request_async(),
                loop=self.event_loop,
            )
            return future
        else:
            if auth:
                headers['Authorization'] = f"Bearer {self.get_token()}"

            return self._sync_request(method=method, url=url, timeout_pair=timeout_pair,
                                      headers=headers, params=params, payload=payload, data=data, files=files)

    async def _async_request(self, method: str, url: str, timeout_pair: Tuple[float, float],
                             headers: dict, params: dict, payload: dict, data: dict, files: dict) -> Future:

        if not self.session_async:
            self.session_async = aiohttp.ClientSession()
        if not self.event_loop or self.event_loop.is_closed():
            self.event_loop = _get_or_create_event_loop()
        request_id = secrets.token_hex(4)

        try:
            timeout = aiohttp.ClientTimeout(sock_connect=timeout_pair[0], sock_read=timeout_pair[1])
            if method == "GET":
                self.logger.debug(f"GET url={url} params={params} headers={headers} (id={request_id})")
                resp = await self.session_async.get(url, params=params, headers=headers, timeout=timeout)
            elif method == "POST":
                if data or files:
                    # multipart/form-data
                    form = aiohttp.FormData()
                    for key, value in data.items():
                        form.add_field(key, value)
                    for filename, content in files.items():
                        form.add_field(filename, content)
                    self.logger.debug(f"POST(form-data) url={url} params={params} "
                                      f"headers={headers} (id={request_id})")
                    resp = await self.session_async.post(url, params=params, data=form, headers=headers,
                                                         timeout=timeout)
                else:
                    # application/json
                    self.logger.debug(f"POST url={url} params={params} json={payload} "
                                      f"headers={headers} (id={request_id})")
                    resp = await self.session_async.post(url, params=params, json=payload, headers=headers,
                                                         timeout=timeout)
            else:
                raise FeishuError(ERRORS.UNSUPPORTED_METHOD,
                                  f"不支持的请求method: {method}, 调用上下文: "
                                  f"url={url}, params={params}, payload={payload} "
                                  f"data={data} files.keys={files.keys()}")
        except aiohttp.ClientError as e:
            raise FeishuError(ERRORS.FAILED_TO_ESTABLISH_CONNECTION, f"建立和服务器的请求失败: {e}")

        try:
            result = await resp.json(content_type=None)
        except ValueError:
            raise FeishuError(ERRORS.UNABLE_TO_PARSE_SERVER_RESPONSE, f"服务器返回格式有问题，无法解析成JSON: {resp.text}")

        if result.get("code") != 0:
            raise FeishuError(result.get("code") or ERRORS.UNKNOWN_SERVER_ERROR,
                              result.get("msg") or f"无有效出错信息，返回JSON数据为: {result}")

        self.logger.debug(f"response={result} (id={request_id})")
        return result

    def _sync_request(self, method: str, url: str, timeout_pair: Tuple[float, float],
                      headers: dict, params: dict, payload: dict, data: dict, files: dict) -> dict:
        request_id = secrets.token_hex(4)
        try:
            if method == "GET":
                self.logger.debug(f"GET url={url} params={params} headers={headers} (id={request_id})")
                resp = self.session.get(url, params=params, headers=headers, timeout=timeout_pair)
            elif method == "POST":
                self.logger.debug(f"POST url={url} params={params} json={payload} data={data} "
                                  f"files.keys={files.keys()} headers={headers} (id={request_id})")
                resp = self.session.post(url, params=params, json=payload, data=data, files=files,
                                         headers=headers, timeout=timeout_pair)
            else:
                raise FeishuError(ERRORS.UNSUPPORTED_METHOD,
                                  f"不支持的请求method: {method}, 调用上下文: "
                                  f"params={params}, payload={payload}")
        except requests.exceptions.RequestException as e:
            raise FeishuError(ERRORS.FAILED_TO_ESTABLISH_CONNECTION, f"建立和服务器的请求失败: {e}")

        try:
            result = resp.json()
        except ValueError:
            raise FeishuError(ERRORS.UNABLE_TO_PARSE_SERVER_RESPONSE, f"服务器返回格式有问题，无法解析成JSON: {resp.text}")

        if result.get("code") != 0:
            raise FeishuError(result.get("code") or ERRORS.UNKNOWN_SERVER_ERROR,
                              result.get("msg") or f"无有效出错信息，返回JSON数据为: {result}")

        self.logger.debug(f"response={result} (id={request_id})")
        return result

    def fetch(self, url: str, params: dict = {}, data: dict = {}, json: dict = {},
              headers: dict = {}, method: str = "GET", timeout: Union[float, tuple] = 2) \
            -> Union[bytes, Future]:
        """简易sync/asyncHTTP请求Adapter

        Usage::

        >>> def api_sync(self, url):
        ...     resp = self.fetch(url)

        >>> async def api_async(self, url):
        ...     resp = await self.fetch(url) # 注意, self.run_async必须为True

        或者更好一点，同时支持sync/async

        >>> @allow_async_call
        ... def api_both(self, url):
        ...     resp = self.fetch(url)
        """
        if self.closed:
            raise FeishuError(ERRORS.CLIENT_CLOSED, "client对象已被关闭")

        if self.run_async:
            if not self.session_async:
                self.session_async = aiohttp.ClientSession()
            if not self.event_loop or self.event_loop.is_closed():
                self.event_loop = _get_or_create_event_loop()

            async def async_fetch():
                if data:
                    resp = await self.session_async.request(method=method, url=url, params=params, data=data,
                                                            headers=headers, timeout=timeout)
                else:
                    resp = await self.session_async.request(method=method, url=url, params=params, json=json,
                                                            headers=headers, timeout=timeout)
                return await resp.read()

            return asyncio.ensure_future(
                async_fetch(),
                loop=self.event_loop
            )
        else:
            if data:
                resp = self.session.request(method=method, url=url, params=params, data=data,
                                            headers=headers, timeout=timeout)
            else:
                resp = self.session.request(method=method, url=url, params=params, json=json,
                                            headers=headers, timeout=timeout)
            return resp.content

    async def close(self):
        """不关闭一下aiohttp会发warning有点烦, 强迫症适用"""
        if not self.closed and self.session_async:
            self.session_async.close()
            self.closed = True



#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author:lenovo
@file: setup.py
@time: 2022/8/15  15:35
"""
import setuptools

# 自动读取readme
with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()
setuptools.setup(
    name="django_feishu_sdk",  # Replace with your own username
    version="0.1.6",
    author="lpf_andr",
    author_email="lpf_andr@163.com",
    description="A feishu api package for django",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/lpfandr/django_feishu_sdk",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=['cryptography', 'pydantic', 'aiohttp', 'requests', 'pycryptodome'],  # 安装所需要的库
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""飞书事件类型声明

参考订阅事件文档 https://open.feishu.cn/document/ukTMukTMukTM/uUTNz4SN1MjL1UzM
"""
from enum import Enum
from typing import Union
import hashlib
import abc
from pydantic import BaseModel

from ...utils import InvalidEventException, Dict2Obj


class EventType(str, Enum):
    APP_OPEN = "app_open"
    APP_STATUS_CHANGE = "app_status_change"
    ORDER_PAID = "order_paid"
    APP_TICKET = "app_ticket"
    APP_UNINSTALLED = "app_uninstalled"

    USER_ADD = "user_add"
    DEPT_ADD = "dept_add"
    USER_STATUS_CHANGE = "user_status_change"
    CONTACT_SCOPE_CHANGE = "contact_scope_change"

    ADD_BOT = "add_bot"
    REMOVE_BOT = "remove_bot"
    P2P_CHAT_CREATE = "p2p_chat_create"
    MESSAGE = "message"
    MESSAGE_READ = "message_read"

    CHAT_DISBAND = "chat_disband"
    GROUP_SETTING_UPDATE = "group_setting_update"

    LEAVE_APPROVAL = "leave_approval"
    LEAVE_APPROVAL_V2 = "leave_approvalV2"
    WORK_APPROVAL = "work_approval"
    SHIFT_APPROVAL = "shift_approval"
    REMEDY_APPROVAL = "remedy_approval"
    TRIP_APPROVAL = "trip_approval"
    OUT_APPROVAL = "out_approval"

    EVENT_REPLY = "event_reply"

    ADD_USER_TO_CHAT = "add_user_to_chat"
    REMOVE_USER_FROM_CHAT = "remove_user_from_chat"
    REVOKE_ADD_USER_FROM_CHAT = "revoke_add_user_from_chat"


class AppStatus(str, Enum):
    START_BY_TENANT = "start_by_tenant"
    STOP_BY_TENANT = "stop_by_tenant"
    STOP_BY_PLATFORM = "stop_by_platform"


class BuyType(str, Enum):
    BUY = "buy"
    UPGRADE = "upgrade"
    RENEW = "renew"


class PricePlanType(str, Enum):
    PER_SEAT_PER_MONTH = "per_seat_per_month"


class UserChatEventType(str, Enum):
    ADD_USER_TO_CHAT = "add_user_to_chat"
    REMOVE_USER_FROM_CHAT = "remove_user_from_chat"
    REVOKE_ADD_USER_FROM_CHAT = "revoke_add_user_from_chat"


class EventHeaderContent(BaseModel):
    event_id: str
    event_type: str
    create_time: str
    token: str
    app_id: str
    tenant_key: str


class EventContent(object):
    """事件回调内容 v2.0
    参考：
    https://open.feishu.cn/document/uAjLw4CM/ukTMukTMukTM/reference/im-v1/message/events/receive
    Args:
        schema 2.0 事件模式(版本)
        header 事件头
        event 事件
    """

    schema: str = ''
    header: Union[dict, EventHeaderContent]
    event: Union[dict]

    def __init__(self, request, dict_data, token, encrypt_key):
        header = dict_data.get("header")
        event = dict_data.get("event")
        if header is None or event is None:
            raise InvalidEventException("request is not callback event(v2)")
        self.__request = request
        self.schema = dict_data.get("schema")
        self.header = Dict2Obj(header)
        self.event = Dict2Obj(event)
        self._validate(token, encrypt_key)

    def _validate(self, token, encrypt_key):
        if self.header.token != token:
            raise InvalidEventException("invalid token")
        timestamp = self.__request.headers.get("X-Lark-Request-Timestamp")
        nonce = self.__request.headers.get("X-Lark-Request-Nonce")
        signature = self.__request.headers.get("X-Lark-Signature")
        body = self.__request.body
        bytes_b1 = (timestamp + nonce + encrypt_key).encode("utf-8")
        bytes_b = bytes_b1 + body
        h = hashlib.sha256(bytes_b)
        if signature != h.hexdigest():
            raise InvalidEventException("invalid signature in event")

    @abc.abstractmethod
    def event_type(self):
        return self.header.event_type
